# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A organization client includes methods needed for connectors.
"""
import logging
from enum import Enum
from unify.properties import Properties, ClusterSetting
from unify.orgadmin import OrgAdmin
from unify.sources import Sources
from unify.templates import Templates
from unify.graph import Graph
from unify.connectors import ConnStatus, LogType, ArtifactType, Connectors

import csv # TODO discuss about how to find templates updated

class DatasetOperation(Enum):
    """
    Operation class represent types of dataset operations
    """
    UPDATE = 1
    APPEND = 2

class OrganizationClient:
    """
    Common connector class
    """
    def __init__(self, user_name, password, org_id, cluster, connector_params):
        self.user_name = user_name
        self.password = password
        self.org_id = org_id
        self.cluster = cluster
        self.connector_params = connector_params

        self.app_name = 'imc'
        self.props = Properties(ClusterSetting.MEMORY)
        self.props.store_cluster(user_name, password, cluster, self.app_name)
        try:
            org_admin = OrgAdmin(self.app_name, self.props)
            token = org_admin.auth_token(
                query_params=connector_params
            )
            self.props.set_auth_token(
                token=token,
                cluster=self.app_name
            )
        except Exception as error:
            logging.error('Failed to login. %s', error)
            raise error

        try:
            self.sources = Sources(self.app_name, self.props)
            self.templates = Templates(self.app_name, self.props)
            self.connectors = Connectors(self.app_name, self.props)
        except Exception as error:
            logging.error('Failed to get organization sources client.')
            raise error

        try:
            self.templates = Templates(self.app_name, self.props)
        except Exception as error:
            logging.error('Failed to get organization templates client.')
            raise error

        try:
            self.graph = Graph(self.app_name, self.props)
        except Exception as error:
            logging.error('Failed to get organization graph client.')
            raise error

    def track_status(func):
        """
        Method used to wrap all other methods to be able to track status of the connector
        """
        def tracking_status(self, *args, **kwargs):
            try:
                self.connectors.update_connector_status(
                    self.org_id,
                    connector_id=self.connector_params['connector_id'],
                    status=ConnStatus.GOOD
                )
                result = func(self, *args, **kwargs)
            except Exception as err:
                self.connectors.update_connector_status(
                    self.org_id,
                    connector_id=self.connector_params['connector_id'],
                    status=ConnStatus.ERROR
                )
                raise
            return result
        return tracking_status

    @track_status
    def list_datasets(self):
        """
        Retrieve all datasets.
        """
        return self.sources.get_sources(org_id=self.org_id)

    @track_status
    def list_datasets_by_labels(self, labels):
        """
        Retrieve all datasets by labels.

        :type labels: list of strings
        :param labels: list of dataset labels. Example ["label1", "label2"]
        """
        return self.sources.get_sources_by_labels(org_id=self.org_id, facets=labels)

    @track_status
    def get_dataset(self, dataset_id):
        """
        Retrieve dataset contents.

        :type dataset_id: string or dict
        :param dataset_id: id of dataset
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        return self.sources.download_dataset_content(org_id=self.org_id, dataset_id=dataset_id_str)

    @track_status
    def create_dataset(self, name, dataset_csv):
        """
        Create a new dataset.

        :type name: string
        :param name: Name of dataset
        :type dataset_csv: string
        :param dataset_csv: Content to upload
        """
        dataset_info =  self.sources.create_api_data_set_with_content(
            name,
            self.org_id,
            content=dataset_csv
        )
        self.check_or_set_artifact(dataset_info['data_set_id'], ArtifactType.DATASET, name)
        return dataset_info

    @track_status
    def update_dataset(self, dataset_csv, dataset_name=None, dataset_id=None):
        """
        Update a dataset. If dataset does not exist, create a new dataset.

        :type dataset_csv: string
        :param dataset_csv: Content to upload
        :type dataset_name: string
        :param dataset_name: Name of dataset
        :type dataset_id: string or dict
        :param dataset_id: Existing dataset id
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        list_datasets = self.list_datasets()
        existing_datasets = [dataset.get("id", {}).get('id') for dataset in list_datasets]
        if dataset_id_str is None or dataset_id_str not in existing_datasets:
            response = self.create_dataset(name=dataset_name, dataset_csv=dataset_csv)
            dataset_id_str = response.get("data_set_id")
        else:
            self.truncate_dataset(dataset_id=dataset_id_str)
            self.append_dataset(dataset_id=dataset_id_str, dataset_csv=dataset_csv)
            dataset_name = [x['name'] for x in list_datasets if x['id']['id']==dataset_id_str][0]

        self.check_or_set_artifact(dataset_id_str, ArtifactType.DATASET, dataset_name)

        return dataset_id_str

    @track_status
    def truncate_dataset(self, dataset_id):
        """
        Truncate a dataset.

        :type dataset_id: string or dict
        :param dataset_id: Existing dataset id
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        result = self.sources.truncate_data_set(org_id=self.org_id, data_set_id=dataset_id_str)
        dataset_name = [x['name'] for x in self.list_datasets() if x['id']['id']==dataset_id_str][0]
        self.check_or_set_artifact(dataset_id_str, ArtifactType.DATASET, dataset_name)
        return result

    @track_status
    def append_dataset(self, dataset_id, dataset_csv):
        """
        Append a dataset.

        :type dataset_csv: string
        :param dataset_csv: Content to upload
        :type dataset_id: string or dict
        :param dataset_id: Existing dataset id
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        result = self.sources.append_dataset(
            org_id=self.org_id,
            data_set_id=dataset_id_str,
            content=dataset_csv)
        dataset_name = [x['name'] for x in self.list_datasets() if x['id']['id']==dataset_id_str][0]
        self.check_or_set_artifact(dataset_id_str, ArtifactType.DATASET, dataset_name)
        return result

    @track_status
    def update_dataset_labels(self, dataset_id, labels):
        """
        Updates labels for a dataset.

        :type dataset_id: string or dict
        :param dataset_id: Existing dataset id
        :type labels: dict
        :param labels: Labels
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        result = self.sources.label(self.org_id, dataset_id_str, labels)
        dataset_name = [x['name'] for x in self.list_datasets() if x['id']['id']==dataset_id_str][0]
        self.check_or_set_artifact(dataset_id_str, ArtifactType.DATASET, dataset_name)
        return result

    @track_status
    def operate_dataset(
        self,
        dataset_csv,
        dataset_id=None,
        dataset_name=None,
        operation=DatasetOperation.UPDATE
    ):
        """
        Operate dataset on given dataset_id. If dataset_id is not given, create a dataset first.
        If dataset is not valid, throw an error.

        :type dataset_csv: string
        :param dataset_csv: Content to upload
        :type dataset_name: string
        :param dataset_name: Name of dataset
        :type dataset_id: string or dict
        :param dataset_id: Existing dataset id
        :type operation: Enum
        :param operation: Operation on dataset, update or append
        """
        dataset_id_str = self.resolve_dataset_id(dataset_id)
        if dataset_id_str is None:
            dataset_id_str = self.create_dataset(dataset_name, dataset_csv)['data_set_id']
            logging.info('create dataset %s id %s to cluster: %s, organization: %d',
                dataset_name, dataset_id_str, self.cluster, self.org_id)
        else:
            # validate dataset_id
            existing_datasets = [
                dataset.get("id", {}).get('id') for dataset in self.list_datasets()
            ]
            if dataset_id_str not in existing_datasets:
                raise Exception(f'Given dataset id {dataset_id_str} does not exists.')
            if operation is DatasetOperation.UPDATE:
                self.truncate_dataset(dataset_id=dataset_id_str)
                self.append_dataset(dataset_id=dataset_id_str, dataset_csv=dataset_csv)
                logging.info('update dataset %s  to cluster: %s, organization: %d',
                    dataset_id_str, self.cluster, self.org_id)
            elif operation is DatasetOperation.APPEND:
                self.append_dataset(dataset_id_str, dataset_csv)
                logging.info('append dataset %s to cluster: %s, organization: %d',
                    dataset_id_str, self.cluster, self.org_id)
        return dataset_id_str


    @track_status
    def upload_template(self, template_csv):
        """
        Upload asset template definition csv.

        :type template_csv: string
        :param template_csv: asset template definition in csv format
        """
        result = self.templates.upload_string_content_file(self.org_id, template_csv)
        all_templates = [(x['id'], x['name']) for x in self.list_templates()]
        templates_updated = {x['Template Name'] for x in csv.DictReader(template_csv.splitlines())}
        for template in all_templates:
            if template[1] not in templates_updated:
                continue
            self.check_or_set_artifact(template[0], ArtifactType.TEMPLATE, template[1])
        
        return result

    @track_status
    def upload_template_configuration(self, config_csv):
        """
        Upload asset template configuration parameter definition csv.

        :type config_csv: string
        :param config_csv: asset template configuration parameter definition in csv format
        """
        result = self.templates.upload_config_with_content(self.org_id, config_csv)
        all_templates = [(x['id'], x['name']) for x in self.list_templates()]
        templates_updated = {x['template'] for x in csv.DictReader(config_csv.splitlines())}
        for template in all_templates:
            if template[1] not in templates_updated:
                continue
            self.check_or_set_artifact(template[0], ArtifactType.TEMPLATE, template[1])
        return result

    @track_status
    def update_template_categories(self, template_id, template_name, version, categories):
        """
        Update categories to an asset template.

        :type template_id: string
        :param template_id: Existing template id
        :type template_name: string
        :param template_name: Existing template name
        :type version: string
        :param version: Existing template id
        :type categories: dict
        :param categories: Labels
        """
        result = self.templates.category(
            self.org_id,
            template_id=template_id,
            template_name=template_name,
            version=version,
            categories=categories
        )
        self.check_or_set_artifact(template_id, ArtifactType.TEMPLATE, template_name)
        return result

    @track_status
    def list_templates(self):
        """
        Retrieve all asset templates
        """
        return self.templates.list_asset_templates(self.org_id)

    @track_status
    def list_graphs(self):
        """
        Retrieve all graphs.
        """
        return self.graph.get_graphs_list(org_id=self.org_id)

    @track_status
    def get_graph(self, graph_id, query=None):
        """
        Retrieve graph.

        :type graph_id: string
        :param graph_id: id of graph
        :type query: string
        :param query: graph cypher query
        """
        if query is None:
            query = "MATCH (source)-[relationship]->(destination) RETURN source, relationship, " \
                "destination"

        return self.graph.query_graph(org_id=self.org_id, graph=graph_id, query=query)

    @staticmethod
    def resolve_dataset_id(dataset_id):
        if dataset_id is None:
            return None
        if isinstance(dataset_id, dict):
            return dataset_id['id']
        if isinstance(dataset_id, str):
            return dataset_id
        raise ValueError(f'dataset_id {dataset_id} is not a valid value')

    def check_or_set_artifact(self, artifact_id, artifact_type, artifact_name):
        """
        Checks if the artifact is already created otherwise it will create it

        :type artifact_id: str
        :param artifact_id: Id of the artifact to be checked or created
        :type artifact_type: str
        :param artifact_type: Type of the artifact to set in case it is not created yet
        :type artifact_name: str
        :param artifact_name: Artifact name to use in case it is not created yet
        """
        artifact_id = str(artifact_id)
        artifacts = self.connectors.get_artifacts(
            self.org_id,
            connector_id=self.connector_params['connector_id']
        )
        if not any(x for x in artifacts if x['artifactId']==artifact_id):
            self.connectors.create_artifact(
                self.org_id,
                connector_id=self.connector_params['connector_id'],
                artifact_type=artifact_type,
                artifact_name=artifact_name,
                artifact_id=artifact_id
            )


# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Abstract class construction to handle the schedule and other triggers
"""


from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from time import sleep
from abc import ABC, abstractmethod
import logging
import json
import random
import string

from enum import Enum

from croniter import croniter
import keyring

class Status(Enum):
    '''
    Enumeration used to check the status of both connector and scheduler
    '''
    IDLE = 0
    RUNNING = 1

class ConnectorHandler(ABC):
    '''
    Abstract class to handle connectors in an on-premise environment with triggers like a schedule
    '''
    def __init__(self, name, config=None):
        """
        Instantiate a new ConnectorHandler class

        :type name: string
        :param name: Name of the Connector Handler, this name will be used to save/load a json file
            with the configuration of the connector, and save logs
        :type config: Union[dict, str, pathlib.Path]
        :param config: Configuration used to setup the connector hanlder, it can be the dict with
            the data or a path to the json file, if is not provided the handler will serch the
            config at the default path and if is not finded, will request you the values via
            terminal input
        """

        self.name = name

        # logs
        path = Path.home().expanduser() / 'Element Unify/Connectors/logs'
        path.mkdir(parents=True, exist_ok=True)
        hdl = logging.FileHandler(path / f'{self.name}.log')
        fmt = logging.Formatter('%(asctime)s - [%(levelname)s] - %(name)s: %(message)s')
        hdl.setFormatter(fmt)
        self.log = logging.getLogger(self.__class__.__name__)
        self.log.addHandler(hdl)
        self.log.setLevel(logging.INFO)
        self.log.info('Saving logs into %s', hdl.baseFilename)

        # Handler variables
        self.config, self.path = self.__get_config(config)
        # self.executions_log = OrderedDict()


        # Trigger variables (schedule only at the moment)
        self.__schedule_status = Status.IDLE
        self.schedule_rule = self.config['schedule']['rule']
        self.scheduler = None # Will store the croniter object with scheduler rule
        self.schedule_executor = ThreadPoolExecutor(1)
        self.schedule_obj = None # Will store the Future object where scheduler is running

        # Execution variables
        self.__connector_status = Status.IDLE
        self.connector = None # This will be instantiated once the execution started
        self.connector_executor = ThreadPoolExecutor(1)
        self.connector_obj = None # Will store the Future object where connector is running

    def start_manual_execution(self):
        """
        Start a thread that will execute the connector behind
        """
        if self.__connector_status == Status.IDLE:
            self.refresh_config()
            self.connector_obj = \
                self.connector_executor.submit(self.__process)
        else:
            self.log.warning(
                'There is already a connector execution in background. SKIPPING execution'
            )

    def __process(self):
        self.__connector_status = Status.RUNNING
        self.log.info(
            'Starting %s running in background, connector status: %s',
            self.connector.__class__.__name__,
            self.__connector_status,
        )
        try:
            self.connector = self.set_connector(self.config)
            self.process(self.config)
            self.log.info('%s excecution finished successfully', self.connector.__class__.__name__)
        except Exception as err:
            self.log.error('An error ocurred during connector execution: %s', err)
            raise
        finally:
            self.__connector_status = Status.IDLE

    def start_trigger(self):
        """
        Start the schedule in a separate thread according with the cron expression provided, this
        will execute the connector once the next datetime provided by the cron expression is met
        """
        if self.__schedule_status == Status.IDLE:
            self.log.info(
                'Startng schedule execution in background, with rule "%s"',
                self.schedule_rule
            )
            self.scheduler = croniter(self.schedule_rule, datetime.now())
            self.log.info('Next execution scheduled at %s', self.scheduler.get_next(datetime))
            self.__schedule_status = Status.RUNNING
            self.schedule_obj = \
                self.schedule_executor.submit(self.__start_schedule)
        else:
            self.log.warning(
                'There is already a schedule execution in background, SKIPPING command'
            )

    def __start_schedule(self):
        while self.__schedule_status == Status.RUNNING:
            scheduled_time = self.scheduler.get_current(datetime)
            if datetime.now() >= scheduled_time:
                self.log.info(
                    'Starting schedule excecition of %s at %s. Next execution scheduled at %s',
                    scheduled_time,
                    datetime.now(),
                    self.scheduler.get_next(datetime)
                )
                self.start_manual_execution()

            sleep(0.25)
        self.log.info('Scheduler was paused, exiting programed scheduled executions')

    def stop(self):
        """
        Stop the scheduler
        """
        self.__schedule_status = Status.IDLE

    def scheduler_is_running(self):
        """
        Returns True if the scheduler is running, False otherwise
        """
        return self.__schedule_status == Status.RUNNING

    def connector_is_running(self):
        """
        Returns True is the connector is executing at the moment, False otherwise like waiting for
        the next execution
        """
        return self.__connector_status == Status.RUNNING

    def status(self):
        """
        Returns a dict with the current status of both scheduler and connector.
        """
        return {
            'Scheduler': self.__schedule_status.name,
            'Connector': self.__connector_status.name,
        }

    @staticmethod
    def set_secret(secret_name, secret_value):
        """
        Method used to safely store the sensible data into the OS keyring service

        :type secret_name: str
        :param secret_name: Name of the secret used later to retrieve the secret value
        :type secret_value: str
        :param secret_value: Value to be stored as a secret
        """
        keyring.set_password('system', f'{secret_name}', secret_value)

    @staticmethod
    def get_secret(secret_name):
        """
        Retrieves the secret value according with the name provided

        :type secret_name: str
        :param secret_name: Name of the secret used to retrieve the secret value
        """
        return keyring.get_password('system', f'{secret_name}')

    @staticmethod
    def rnd_str(length):
        """
        Create a random alphanumeric string of length n

        :type n: int
        :param n: length of the string
        """
        return ''.join(random.choice(string.ascii_letters + string.digits) for x in range(length))

    def __get_config(self, config=None):
        # TODO modify this method when its ready to get the config online

        default_path = f'~/Element Unify/Connectors/config/{self.name}.json'
        Path(default_path).expanduser().parent.mkdir(parents=True, exist_ok=True)

        # If there is not config,
        if config is None:
            # search by name
            if Path(default_path).expanduser().exists():
                return self.get_config(default_path), default_path

            # create it by inputs
            config = self.input_config()

        # once we have the config, store it in the home directory
        if isinstance(config, dict):
            self.__save_config(default_path, config)
            return config, default_path

        # if the config is a path, open it
        return self.get_config(config), config

    def refresh_config(self):
        """
        Reloads the config from the json file
        """
        self.config = self.get_config(self.path)

    def __save_config(self, path, config):
        self.log.info('Saving config in: %s', path)
        data = {
            'type': f'{self.__class__.__name__}',
            'config': config,
        }
        with open(Path(path).expanduser(), 'w') as txt:
            json.dump(data, txt, indent=4)

    def get_config(self, path):
        """
        Returns the config dict stored in the json file specified in path parameter

        :type path: Union[str, pathlib.Path]
        :param path: json filename to be loaded
        """
        # TODO think about getting the config from a Yaml with any syntax PyYAML
        self.log.info('Loading config from: %s', path)
        with open(Path(path).expanduser()) as txt:
            config = json.load(txt)
        return config['config']

    @abstractmethod
    def input_config(self):
        """
        Abstract method to implement the config dict for each connector type, this method should
        return a dict with all the data required to instantiate the connector and execute it.
        """

    @abstractmethod
    def set_connector(self, config):
        """
        Abstract method to implement the creation of the connector, should have a parameter called
        config that will be the dict returned by the method input_config. Should return a
        connector instance. This instance will be stored as self.connector variable.

        :type config: dict
        :param config: dict returned by input_config method
        """

    @abstractmethod
    def process(self, config):
        """
        Abstract method to implement the execution of the connector, it should have a parammeter
        called config that will be the dict returned by the method input_config.

        :type config: dict
        :param config: dict returned by input_config method
        """
